import copy
import os

import kerastuner
import tensorflow as tf
from tensorflow.keras import callbacks as tf_callbacks
from tensorflow.keras.layers.experimental import preprocessing
from tensorflow.python.util import nest

from autokeras.utils import utils


class AutoTuner(kerastuner.engine.tuner.Tuner):
    """A Tuner class based on KerasTuner for AutoKeras.

    Different from KerasTuner's Tuner class. AutoTuner's not only tunes the
    Hypermodel which can be directly built into a Keras model, but also the
    preprocessors. Therefore, a HyperGraph stores the overall search space containing
    both the Preprocessors and Hypermodel. For every trial, the HyperGraph build the
    PreprocessGraph and KerasGraph with the provided HyperParameters.

    The AutoTuner uses EarlyStopping for acceleration during the search and fully
    train the model with full epochs and with both training and validation data.
    The fully trained model is the best model to be used by AutoModel.

    # Arguments
        preprocessors: An instance or list of `Preprocessor` objects corresponding to
            each AutoModel input, to preprocess a `tf.data.Dataset` before passing it
            to the model. Defaults to None (no external preprocessing).
        **kwargs: The args supported by KerasTuner.
    """

    def __init__(self,
                 oracle,
                 hypermodel,
                 preprocessors=None,
                 **kwargs):
        # Initialize before super() for reload to work.
        self._finished = False
        super().__init__(oracle, hypermodel, **kwargs)
        self.preprocessors = nest.flatten(preprocessors)
        # Save or load the HyperModel.
        self.hypermodel.hypermodel.save(os.path.join(self.project_dir, 'graph'))

    # Override the function to prevent building the model during initialization.
    def _populate_initial_space(self):
        pass

    def get_best_model(self):
        model = super().get_best_models()[0]
        model.load_weights(self.best_model_path)
        return model

    def _on_train_begin(self, model, hp, x, *args, **kwargs):
        """Adapt the preprocessing layers and tune the fit arguments."""
        self.adapt(model, x)

    @staticmethod
    def adapt(model, dataset):
        """Adapt the preprocessing layers in the model."""
        # Currently, only support using the original dataset to adapt all the
        # preprocessing layers before the first non-preprocessing layer.
        # TODO: Use PreprocessingStage for preprocessing layers adapt.
        # TODO: Use Keras Tuner for preprocessing layers adapt.
        x = dataset.map(lambda x, y: x)

        def get_output_layer(tensor):
            tensor = nest.flatten(tensor)[0]
            for layer in model.layers:
                if isinstance(layer, tf.keras.layers.InputLayer):
                    continue
                input_node = nest.flatten(layer.input)[0]
                if input_node is tensor:
                    return layer
            return None

        for index, input_node in enumerate(nest.flatten(model.input)):
            temp_x = x.map(lambda *args: nest.flatten(args)[index])
            layer = get_output_layer(input_node)
            while isinstance(layer, preprocessing.PreprocessingLayer):
                layer.adapt(temp_x)
                layer = get_output_layer(layer.output)
        return model

    def search(self,
               epochs=None,
               callbacks=None,
               fit_on_val_data=False,
               **fit_kwargs):
        """Search for the best HyperParameters.

        If there is not early-stopping in the callbacks, the early-stopping callback
        is injected to accelerate the search process. At the end of the search, the
        best model will be fully trained with the specified number of epochs.

        # Arguments
            callbacks: A list of callback functions. Defaults to None.
            fit_on_val_data: Boolean. Use the training set and validation set for the
                final fit of the best model.
        """
        if self._finished:
            return

        if callbacks is None:
            callbacks = []

        # Insert early-stopping for adaptive number of epochs.
        epochs_provided = True
        if epochs is None:
            epochs_provided = False
            epochs = 1000
            if not utils.contain_instance(callbacks, tf_callbacks.EarlyStopping):
                callbacks.append(tf_callbacks.EarlyStopping(patience=10))

        # Insert early-stopping for acceleration.
        early_stopping_inserted = False
        new_callbacks = self._deepcopy_callbacks(callbacks)
        if not utils.contain_instance(callbacks, tf_callbacks.EarlyStopping):
            early_stopping_inserted = True
            new_callbacks.append(tf_callbacks.EarlyStopping(patience=10))

        # Populate initial search space.
        hp = self.oracle.get_space()
        self.hypermodel.build(hp)
        self.oracle.update_space(hp)

        super().search(epochs=epochs, callbacks=new_callbacks, **fit_kwargs)

        # Train the best model use validation data.
        # Train the best model with enought number of epochs.
        if fit_on_val_data or early_stopping_inserted:
            copied_fit_kwargs = copy.copy(fit_kwargs)

            # Remove early-stopping since no validation data.
            # Remove early-stopping since it is inserted.
            copied_fit_kwargs['callbacks'] = self._remove_early_stopping(callbacks)

            # Decide the number of epochs.
            copied_fit_kwargs['epochs'] = epochs
            if not epochs_provided:
                copied_fit_kwargs['epochs'] = self._get_best_trial_epochs()

            # Concatenate training and validation data.
            if fit_on_val_data:
                copied_fit_kwargs['x'] = copied_fit_kwargs['x'].concatenate(
                    fit_kwargs['validation_data'])
                copied_fit_kwargs.pop('validation_data')

            model = self.final_fit(**copied_fit_kwargs)
        else:
            model = self.get_best_models()[0]

        model.save_weights(self.best_model_path)
        self._finished = True

    def get_state(self):
        state = super().get_state()
        state.update({
            'finished': self._finished,
        })
        return state

    def set_state(self, state):
        super().set_state(state)
        self._finished = state.get('finished')

    @staticmethod
    def _remove_early_stopping(callbacks):
        return [copy.deepcopy(callbacks)
                for callback in callbacks
                if not isinstance(callback, tf_callbacks.EarlyStopping)]

    def _get_best_trial_epochs(self):
        best_trial = self.oracle.get_best_trials(1)[0]
        return len(best_trial.metrics.metrics['val_loss']._observations)

    def final_fit(self, x=None, **fit_kwargs):
        best_trial = self.oracle.get_best_trials(1)[0]
        best_hp = best_trial.hyperparameters
        model = self.hypermodel.build(best_hp)
        self.adapt(model, x)
        model.fit(x, **fit_kwargs)
        return model

    @property
    def best_model_path(self):
        return os.path.join(self.project_dir, 'best_model')

    @property
    def objective(self):
        return self.tuner.objective

    @property
    def max_trials(self):
        return self.oracle.max_trials



from unittest import mock

import kerastuner

from autokeras import graph as graph_module
from autokeras.tuners import greedy
from tests import utils


def test_greedy_oracle_state_hypermodel_is_graph():
    oracle = greedy.GreedyOracle(
        hypermodel=utils.build_graph(),
        objective='val_loss',
    )
    oracle.set_state(oracle.get_state())
    assert isinstance(oracle.hypermodel, graph_module.Graph)


def test_greedy_oracle_get_state_update_space_can_run():
    oracle = greedy.GreedyOracle(
        hypermodel=utils.build_graph(),
        objective='val_loss',
    )
    oracle.set_state(oracle.get_state())
    hp = kerastuner.HyperParameters()
    hp.Boolean('test')
    oracle.update_space(hp)


@mock.patch('autokeras.tuners.greedy.GreedyOracle.get_best_trials')
def test_greedy_oracle(fn):
    oracle = greedy.GreedyOracle(
        hypermodel=utils.build_graph(),
        objective='val_loss',
    )
    trial = mock.Mock()
    hp = kerastuner.HyperParameters()
    trial.hyperparameters = hp
    fn.return_value = [trial]

    oracle.update_space(hp)
    for i in range(2000):
        oracle._populate_space(str(i))

    assert 'optimizer' in oracle._hp_names[greedy.GreedyOracle.OPT]
    assert 'classification_head_1/dropout' in oracle._hp_names[
        greedy.GreedyOracle.ARCH]
    assert 'image_block_1/block_type' in oracle._hp_names[
        greedy.GreedyOracle.HYPER]



from distutils.core import setup
from pathlib import Path

from setuptools import find_packages

this_file = Path(__file__).resolve()
readme = this_file.parent / 'README.md'

setup(
    name='autokeras',
    version='1.0.4',
    description='AutoML for deep learning',
    package_data={'': ['README.md']},
    long_description=readme.read_text(encoding='utf-8'),
    long_description_content_type='text/markdown',
    author='Data Analytics at Texas A&M (DATA) Lab, Keras Team',
    author_email='jhfjhfj1@gmail.com',
    url='http://autokeras.com',
    download_url='https://github.com/keras-team/autokeras/archive/1.0.4.tar.gz',
    keywords=['AutoML', 'Keras'],
    install_requires=[
        'packaging',
        'tensorflow>=2.2.0',
        'scikit-learn',
        'numpy',
        'pandas',
    ],
    extras_require={
        'tests': ['pytest>=4.4.0',
                  'flake8',
                  'isort',
                  'pytest-xdist',
                  'pytest-cov',
                  'coverage',
                  'typeguard>=2,<2.10.0',
                  'typedapi>=0.2,<0.3'
                  ],
    },
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        'License :: OSI Approved :: MIT License',
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: Libraries",
    ],
    license="MIT",
    packages=find_packages(exclude=('tests',)),
)



import kerastuner
import tensorflow as tf
from tensorflow.python.util import nest

from autokeras import adapters
from autokeras import blocks
from tests import utils


def test_image_build_return_tensor():
    block = blocks.ImageBlock()

    outputs = block.build(
        kerastuner.HyperParameters(),
        tf.keras.Input(shape=(32, 32, 3), dtype=tf.float32))

    assert len(nest.flatten(outputs)) == 1
    assert isinstance(nest.flatten(outputs)[0], tf.Tensor)


def test_image_deserialize_to_image():
    serialized_block = blocks.serialize(blocks.ImageBlock())

    block = blocks.deserialize(serialized_block)

    assert isinstance(block, blocks.ImageBlock)


def test_image_get_config_has_all_attributes():
    block = blocks.ImageBlock()

    config = block.get_config()

    assert utils.get_func_args(
        blocks.ImageBlock.__init__).issubset(config.keys())


def test_text_build_return_tensor():
    block = blocks.TextBlock()

    outputs = block.build(
        kerastuner.HyperParameters(),
        tf.keras.Input(shape=(1,), dtype=tf.string))

    assert len(nest.flatten(outputs)) == 1
    assert isinstance(nest.flatten(outputs)[0], tf.Tensor)


def test_text_deserialize_to_text():
    serialized_block = blocks.serialize(blocks.TextBlock())

    block = blocks.deserialize(serialized_block)

    assert isinstance(block, blocks.TextBlock)


def test_text_get_config_has_all_attributes():
    block = blocks.TextBlock()

    config = block.get_config()

    assert utils.get_func_args(
        blocks.TextBlock.__init__).issubset(config.keys())


def test_structured_build_return_tensor():
    block = blocks.StructuredDataBlock()
    block.column_names = ['0', '1']
    block.column_types = {'0': adapters.NUMERICAL,
                          '1': adapters.NUMERICAL}

    outputs = block.build(
        kerastuner.HyperParameters(),
        tf.keras.Input(shape=(2,), dtype=tf.string))

    assert len(nest.flatten(outputs)) == 1
    assert isinstance(nest.flatten(outputs)[0], tf.Tensor)


def test_structured_deserialize_to_structured():
    serialized_block = blocks.serialize(blocks.StructuredDataBlock())

    block = blocks.deserialize(serialized_block)

    assert isinstance(block, blocks.StructuredDataBlock)


def test_structured_get_config_has_all_attributes():
    block = blocks.StructuredDataBlock()

    config = block.get_config()

    assert utils.get_func_args(
        blocks.StructuredDataBlock.__init__).issubset(config.keys())


def test_timeseries_build_return_tensor():
    block = blocks.TimeseriesBlock()
    block.column_names = ['0', '1']
    block.column_types = {'0': adapters.NUMERICAL,
                          '1': adapters.NUMERICAL}

    outputs = block.build(
        kerastuner.HyperParameters(),
        tf.keras.Input(shape=(32, 2), dtype=tf.float32))

    assert len(nest.flatten(outputs)) == 1
    assert isinstance(nest.flatten(outputs)[0], tf.Tensor)


def test_timeseries_deserialize_to_timeseries():
    serialized_block = blocks.serialize(blocks.TimeseriesBlock())

    block = blocks.deserialize(serialized_block)

    assert isinstance(block, blocks.TimeseriesBlock)


def test_timeseries_get_config_has_all_attributes():
    block = blocks.TimeseriesBlock()

    config = block.get_config()

    assert utils.get_func_args(
        blocks.TimeseriesBlock.__init__).issubset(config.keys())



import os

import numpy as np
import tensorflow as tf

from autokeras import keras_layers as layer_module


def get_data():
    train = np.array([['a', 'ab', 2.1], ['b', 'bc', 1.0], ['a', 'bc', 'nan']])
    test = np.array([['a', 'ab', 2.1], ['x', 'bc', 1.0], ['a', 'bc', 'nan']])
    y = np.random.rand(3, 1)
    return train, test, y


def test_multi_column_categorical_encoding(tmp_path):
    x_train, x_test, y_train = get_data()
    input_node = tf.keras.Input(shape=(3,), dtype=tf.string)
    layer = layer_module.MultiColumnCategoricalEncoding([
        layer_module.INT,
        layer_module.INT,
        layer_module.NONE,
    ])
    hidden_node = layer(input_node)
    output_node = tf.keras.layers.Dense(1, activation='sigmoid')(hidden_node)
    model = tf.keras.Model(input_node, output_node)
    model.compile(loss='binary_crossentropy', optimizer='adam')
    tf.data.Dataset.zip((
        (tf.data.Dataset.from_tensor_slices(x_train).batch(32),),
        (tf.data.Dataset.from_tensor_slices(np.random.rand(3, 1)).batch(32),),
    ))
    layer.adapt(tf.data.Dataset.from_tensor_slices(x_train).batch(32))

    model.fit(x_train, y_train, epochs=1)

    model2 = tf.keras.Model(input_node, hidden_node)
    result = model2.predict(x_train)
    assert result[0][0] == result[2][0]
    assert result[0][0] != result[1][0]
    assert result[0][1] != result[1][1]
    assert result[0][1] != result[2][1]
    assert result[2][2] == 0

    output = model2.predict(x_test)
    assert output.dtype == np.dtype('float32')


def build_model():
    input_node = tf.keras.Input(shape=(3,), dtype=tf.string)
    layer = layer_module.MultiColumnCategoricalEncoding(encoding=[
        layer_module.INT, layer_module.INT, layer_module.NONE])
    output_node = layer(input_node)
    output_node = tf.keras.layers.Dense(1)(output_node)
    return tf.keras.Model(input_node, output_node), layer


def test_model_save(tmp_path):
    x_train, x_test, y_train = get_data()
    model, layer = build_model()
    layer.adapt(tf.data.Dataset.from_tensor_slices(x_train).batch(32))
    model.compile(optimizer='adam', loss='mse')
    model.fit(x_train, y_train, epochs=1, verbose=False)

    model.save(os.path.join(tmp_path, 'model'))
    model2 = tf.keras.models.load_model(os.path.join(tmp_path, 'model'),
                                        custom_objects=layer_module.CUSTOM_OBJECTS)

    assert np.array_equal(model.predict(x_train), model2.predict(x_train))


def test_weight_save(tmp_path):
    x_train, x_test, y_train = get_data()
    model, layer = build_model()
    layer.adapt(tf.data.Dataset.from_tensor_slices(x_train).batch(32))
    model.compile(optimizer='adam', loss='mse')
    model.fit(x_train, y_train, epochs=1, verbose=False)
    model.save_weights(os.path.join(tmp_path, 'checkpoint'))

    model2, _ = build_model()
    model2.load_weights(os.path.join(tmp_path, 'checkpoint'))

    assert np.array_equal(model.predict(x_train), model2.predict(x_train))



import kerastuner
import numpy as np
import tensorflow as tf

import autokeras as ak
from autokeras import blocks
from autokeras import nodes as input_module
from autokeras.blocks import heads as head_module


def test_two_classes():
    y = np.array(['a', 'a', 'a', 'b'])
    head = head_module.ClassificationHead(name='a')
    adapter = head.get_adapter()
    adapter.fit_transform(y)
    head.config_from_adapter(adapter)
    head.output_shape = (1,)
    head.build(kerastuner.HyperParameters(), input_module.Input(shape=(32,)).build())
    assert head.loss.name == 'binary_crossentropy'


def test_three_classes():
    y = np.array(['a', 'a', 'c', 'b'])
    head = head_module.ClassificationHead(name='a')
    adapter = head.get_adapter()
    adapter.fit_transform(y)
    head.config_from_adapter(adapter)
    assert head.loss.name == 'categorical_crossentropy'


def test_multi_label_loss():
    head = head_module.ClassificationHead(name='a', multi_label=True, num_classes=8)
    head.output_shape = (8,)
    input_node = tf.keras.Input(shape=(5,))
    output_node = head.build(kerastuner.HyperParameters(), input_node)
    model = tf.keras.Model(input_node, output_node)
    assert model.layers[-1].activation.__name__ == 'sigmoid'
    assert head.loss.name == 'binary_crossentropy'


def test_segmentation():
    y = np.array(['a', 'a', 'c', 'b'])
    head = head_module.SegmentationHead(name='a')
    adapter = head.get_adapter()
    adapter.fit_transform(y)
    head.config_from_adapter(adapter)
    input_shape = (64, 64, 21)
    hp = kerastuner.HyperParameters()
    head = blocks.deserialize(blocks.serialize(head))
    head.build(hp, ak.Input(shape=input_shape).build())



import pytest

from autokeras.engine import block as block_module


def test_block_call_raise_inputs_type_error():
    block = block_module.Block()

    with pytest.raises(TypeError) as info:
        block(None)

    assert 'Expect the inputs to block' in str(info.value)



from tensorflow.python.keras.datasets import mnist

import autokeras as ak
from tests import utils


def test_text_and_structured_data(tmp_path):
    # Prepare the data.
    num_instances = 80
    (x_text, y_train), (x_test, y_test) = utils.imdb_raw()
    (x_structured_data, y_train), (x_test, y_test) = utils.dataframe_numpy()

    x_text = x_text[:num_instances]
    x_structured_data = x_structured_data[:num_instances]
    y_classification = utils.generate_one_hot_labels(num_instances=num_instances,
                                                     num_classes=3)
    y_regression = utils.generate_data(num_instances=num_instances, shape=(1,))

    # Build model and train.
    structured_data_input = ak.StructuredDataInput()
    structured_data_output = ak.CategoricalToNumerical()(structured_data_input)
    structured_data_output = ak.DenseBlock()(structured_data_output)

    text_input = ak.TextInput()
    outputs1 = ak.TextToIntSequence()(text_input)
    outputs1 = ak.Embedding()(outputs1)
    outputs1 = ak.ConvBlock(separable=True)(outputs1)
    outputs1 = ak.SpatialReduction()(outputs1)
    outputs2 = ak.TextToNgramVector()(text_input)
    outputs2 = ak.DenseBlock()(outputs2)
    text_output = ak.Merge()((
        outputs1,
        outputs2
    ))

    merged_outputs = ak.Merge()((
        structured_data_output,
        text_output
    ))

    regression_outputs = ak.RegressionHead()(merged_outputs)
    classification_outputs = ak.ClassificationHead()(merged_outputs)
    automodel = ak.AutoModel(
        inputs=[
            text_input,
            structured_data_input
        ],
        directory=tmp_path,
        outputs=[
            regression_outputs,
            classification_outputs
        ],
        max_trials=2,
        tuner=ak.Hyperband,
        seed=utils.SEED)

    automodel.fit(
        (
            x_text,
            x_structured_data
        ),
        (
            y_regression,
            y_classification
        ),
        validation_split=0.2,
        epochs=1)


def test_image_blocks(tmp_path):
    num_instances = 10
    (x_train, y_train), (x_test, y_test) = mnist.load_data()
    x_train = x_train[:num_instances]
    y_regression = utils.generate_data(num_instances=num_instances, shape=(1,))

    input_node = ak.ImageInput()
    output = ak.Normalization()(input_node)
    output = ak.ImageAugmentation()(output)
    outputs1 = ak.ResNetBlock(version='v2')(output)
    outputs2 = ak.XceptionBlock()(output)
    output_node = ak.Merge()((outputs1, outputs2))
    output_node = ak.ClassificationHead()(output_node)

    automodel = ak.AutoModel(inputs=input_node,
                             outputs=output_node,
                             directory=tmp_path,
                             max_trials=1,
                             seed=utils.SEED)

    automodel.fit(x_train, y_regression,
                  validation_data=(x_train, y_regression), epochs=1)



from unittest import mock

import numpy as np
import pandas
import pytest

import autokeras as ak
from tests import utils


def test_raise_error_unknown_str_in_col_type(tmp_path):
    with pytest.raises(ValueError) as info:
        ak.StructuredDataClassifier(
            column_types={'age': 'num', 'parch': 'categorical'},
            directory=tmp_path,
            seed=utils.SEED)

    assert 'Column_types should be either "categorical"' in str(info.value)


@mock.patch('autokeras.AutoModel.fit')
def test_structured_clf_fit_call_auto_model_fit(fit, tmp_path):
    auto_model = ak.StructuredDataClassifier(directory=tmp_path, seed=utils.SEED)

    auto_model.fit(
        x=utils.generate_structured_data(num_instances=100),
        y=utils.generate_one_hot_labels(num_instances=100, num_classes=3))

    assert fit.is_called


@mock.patch('autokeras.AutoModel.fit')
def test_structured_reg_fit_call_auto_model_fit(fit, tmp_path):
    auto_model = ak.StructuredDataRegressor(directory=tmp_path, seed=utils.SEED)

    auto_model.fit(
        x=utils.generate_structured_data(num_instances=100),
        y=utils.generate_data(num_instances=100, shape=(1,)))

    assert fit.is_called


@mock.patch('autokeras.AutoModel.fit')
def test_structured_data_clf_convert_csv_to_df_and_np(fit, tmp_path):
    auto_model = ak.StructuredDataClassifier(directory=tmp_path, seed=utils.SEED)

    auto_model.fit(x=utils.TRAIN_FILE_PATH,
                   y='survived',
                   epochs=2,
                   validation_data=(utils.TEST_FILE_PATH, 'survived'))

    _, kwargs = fit.call_args_list[0]
    assert isinstance(kwargs['x'], pandas.DataFrame)
    assert isinstance(kwargs['y'], np.ndarray)
